#!/usr/bin/env python3
# coding: utf-8

import os
import os.path
import shlex
import subprocess
import sys
import warnings
from glob import glob
from typing import List

from joker.interfaces.filesys import Directory

_warning = """\
joker.interfaces.repos is deprecated, 
please use joker.filesys.git instead.
"""

warnings.warn(
    _warning,
    DeprecationWarning
)
__all__ = []


def printcmd(cmd: List[str], **kwargs):
    kwargs.setdefault('file', sys.stderr)
    print(*[shlex.quote(s) for s in cmd], **kwargs)


class Repository(Directory):
    def __init__(self, path: str):
        dotgit_dir = os.path.join(path, '.git')
        if not os.path.isdir(dotgit_dir):
            raise NotADirectoryError(dotgit_dir)
        super().__init__(path)

    def under_dotgit_dir(self, *paths):
        return self.under('.git', *paths)

    @classmethod
    def find(cls, path: str) -> List['Repository']:
        pattern = os.path.join(path, '*', '.git')
        dotgit_paths = glob(pattern)
        for dotgit_path in dotgit_paths:
            try:
                yield cls(os.path.split(dotgit_path)[0])
            except NotADirectoryError:
                continue

    def pull(self):
        cmd = ['git', 'pull']
        printcmd(cmd)
        subprocess.run(cmd, cwd=self.base_dir)

    def check_command(self, cmd: list):
        sp = subprocess.run(cmd, cwd=self.base_dir, capture_output=True)
        return sp.stdout.decode('utf-8').strip()

    def get_current_branch(self) -> str:
        cmd = ['git', 'rev-parse', '--abbrev-ref', 'HEAD']
        return self.check_command(cmd)

    def get_status_lines(self) -> list:
        cmd = ['git', 'status', '--short']
        return self.check_command(cmd).splitlines()

    def get_commit_info(self) -> dict:
        cmd = ['git', 'log', '-1', "--pretty=%ae%n%h%n%ci%n%s"]
        keys = ['author', 'commit', 'committed_at', 'message']
        lines = self.check_command(cmd).splitlines()
        if len(lines) < len(keys):
            return {}
        info = dict(zip(keys, lines))
        info.update({
            'branch': self.get_current_branch(),
            'status': self.get_status_lines(),
        })
        return info






#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import os.path
import urllib.parse
from typing import Generator

from joker.filesys.archives import (
    Archive, TarArchive, ZipArchive,
)
from joker.filesys.utils import (
    read_as_chunks, compute_checksum, checksum, checksum_hexdigest,
    b64_encode_data_url, b64_encode_local_file,
)

_names = [
    read_as_chunks, compute_checksum, checksum, checksum_hexdigest,
    b64_encode_data_url, b64_encode_local_file,
    Archive, TarArchive, ZipArchive,
]

_warning = """\
joker.interfaces.filesys is deprecated, 
please use joker.filesys instead.
"""

warnings.warn(
    _warning,
    DeprecationWarning
)
__all__ = []


class Directory:
    def __init__(self, base_dir: str):
        self.base_dir = os.path.abspath(base_dir)

    def __repr__(self):
        c = self.__class__.__name__
        return '{}({!r})'.format(c, self.base_dir)

    def under(self, *paths):
        return os.path.join(self.base_dir, *paths)

    def relative_to_base_dir(self, path: str):
        path = os.path.abspath(path)
        return os.path.relpath(path, self.base_dir)

    under_base_dir = under

    def read_as_chunks(self, path: str, length=-1, offset=0, chunksize=65536) \
            -> Generator[bytes, None, None]:
        path = self.under_base_dir(path)
        return read_as_chunks(
            path, length=length,
            offset=offset, chunksize=chunksize,
        )

    def checksum_hexdigest(self, path: str, algo='sha1') -> str:
        path = self.under_base_dir(path)
        return checksum_hexdigest(path, algo=algo)

    def read_as_binary(self, path: str):
        path = self.under_base_dir(path)
        with open(path, 'rb') as fin:
            return fin.read()

    def read_as_base64_data_url(self, path: str):
        path = self.under_base_dir(path)
        return b64_encode_local_file(path)

    def save_as_file(self, path: str, chunks):
        path = self.under_base_dir(path)
        with open(path, 'wb') as fout:
            for chunk in chunks:
                fout.write(chunk)


class MappedDirectory(Directory):
    def __init__(self, base_dir: str, base_url: str):
        super().__init__(base_dir)
        # Note:
        # urllib.parse.urljoin('/a/b', 'c.jpg') => '/a/c.jpg'
        # urllib.parse.urljoin('/a/b/', 'c.jpg') => '/a/b/c.jpg'
        if not base_url.endswith('/'):
            base_url += '/'
        self.base_url = base_url

    def __repr__(self):
        c = self.__class__.__name__
        return '{}({!r}, {!r})'.format(c, self.base_dir, self.base_url)

    def join_url(self, path: str):
        return urllib.parse.urljoin(self.base_url, path)

    def relative_to_base_url(self, url: str):
        base_url_path = urllib.parse.urlparse(self.base_url).path
        url_path = urllib.parse.urlparse(url).path
        return os.path.relpath(url_path, base_url_path)

    def convert_local_path_to_url(self, path: str):
        path = os.path.abspath(path)
        path = self.relative_to_base_dir(path)
        return self.join_url(path)

    def convert_url_to_local_path(self, url: str):
        path = self.relative_to_base_url(url)
        return self.under(path)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import json
import logging
import shlex
import traceback
import urllib.parse
from functools import cached_property
from json import JSONDecodeError
from typing import Union
from urllib.parse import urljoin, urlparse

import requests
from volkanic.errors import TechnicalError
from volkanic.introspect import razor

_logger = logging.getLogger(__name__)


class ResponseDict(dict):
    @property
    def code(self) -> int:
        return self.get('code', 3)

    @property
    def data(self):
        return self.get('data')

    @property
    def message(self):
        return self.get('message', 'OK')


def dump_json_request_to_curl(method: str, url: str, data=None, aslist=False):
    method = method.upper()
    if method == 'GET':
        parts = ['curl', url]
    else:
        parts = [
            'curl', '-X', method, url,
            '-H', 'Content-Type: application/json',
            '-d', json.dumps(razor(data), ensure_ascii=False),
        ]
    if aslist:
        return parts
    parts = [shlex.quote(s) for s in parts]
    return ' '.join(parts)


def _log_bad_response(resp: requests.Response):
    _logger.error(
        'bad response: %s %r',
        resp.status_code, resp.content[:1000]
    )


def _decode_response(resp: requests.Response):
    status = resp.status_code
    if status >= 400:
        raise TechnicalError(f'got response status code {status}')
    try:
        rd = ResponseDict(resp.json())
    except JSONDecodeError:
        raise TechnicalError('cannot decode json')
    if rd.code != 0:
        raise TechnicalError(f'error response ({rd.code})')
    if rd.message:
        _logger.info(rd.message)
    return rd.data


def decode_response(resp: requests.Response):
    try:
        return _decode_response(resp)
    except TechnicalError:
        _log_bad_response(resp)
        raise


class _BaseHTTPClient:
    @staticmethod
    def _check_url(url: str):
        path = urlparse(url).path
        if path == '' or path.endswith('/'):
            return
        raise ValueError('service url path must end with "/"')

    def __init__(self, url: str):
        self._check_url(url)
        self.base_url = url
        c = self.__class__.__name__
        _logger.info('new %s instance, %r', c, url)

    @cached_property
    def session(self):
        return requests.session()


class HTTPClient(_BaseHTTPClient):
    timeout = 30

    def get_url(self, path: str):
        return urllib.parse.urljoin(self.base_url, path)

    @staticmethod
    def _load_json(resp: requests.Response) -> Union[dict, list, None]:
        err = _logger.error
        if resp.status_code != 200:
            err('status_code %s from %s', resp.status_code, resp.url)
            return
        try:
            data = json.loads(resp.text)
        except json.JSONDecodeError:
            traceback.print_exc()
            err('bad json from %s', resp.url)
            return
        if not isinstance(data, dict):
            err('wrong response type (%s) from %s', type(data), resp.url)
        code = data.get('code')
        if data.get('code'):
            err('non-zero code (%s) from %s', code, resp.url)
            return
        return data

    @staticmethod
    def log_request_as_curl(
            method: str, url: str, data=None, level=logging.DEBUG):
        if not _logger.isEnabledFor(level):
            return
        cmd = dump_json_request_to_curl(method, url, data=data)
        _logger.log(level, cmd)

    @staticmethod
    def log_resp_content(
            resp: requests.Response, level=logging.ERROR):
        if not _logger.isEnabledFor(level):
            return
        try:
            _logger.log(level, resp.text[:255])
        except Exception:
            _logger.log(level, resp.content[:255])
            traceback.print_exc()

    def json_request(self, method, path, data=None, **kwargs) \
            -> Union[dict, list, None]:
        url = self.get_url(path)
        method = method.upper()
        kwargs.setdefault('timeout', self.timeout)
        if data is not None:
            kwargs['data'] = json.dumps(data, default=str)
            kwargs['headers'] = {'Content-type': 'application/json'}
        resp = requests.request(method, url, **kwargs)
        resp_data = self._load_json(resp)
        if resp_data is None:
            self.log_resp_content(resp)
            self.log_request_as_curl(method, url, data)
            return
        self.log_request_as_curl(method, url, data, level=logging.DEBUG)
        return resp_data

    def json_get(self, path, **kwargs) -> Union[dict, list, None]:
        assert 'data' not in kwargs
        return self.json_request('GET', path, **kwargs)

    def json_post(self, path, data=None, **kwargs) -> Union[dict, list, None]:
        assert 'data' not in kwargs
        return self.json_request('POST', path, data=data, **kwargs)


# will be removed in 0.2.0
HTTPService = HTTPClient


# deprecated; will be removed in 0.2.0
class HTTPClientInterface:
    def __init__(self, config: dict):
        self._upstream_config = config

    def __getitem__(self, service_name: str) -> HTTPService:
        return HTTPService(self._upstream_config[service_name])

    def get_service_url(self, service_name: str, path: str = None):
        base_url = self._upstream_config[service_name]
        if not path:
            return base_url
        return urljoin(base_url, path)


# deprecated; will be removed in 0.2.0
HTTPServiceInterface = HTTPClientInterface
